@@ -37,7 +37,9 @@ USE OR OTHER DEALINGS IN THE SOFTWARE.
 #endif
 
 #include <errno.h>
+#include <time.h>
 #include <string.h>
+#include <unistd.h>
 
 #include "sna.h"
 #include "sna_reg.h"
@@ -175,6 +177,7 @@ static struct kgem_bo *sna_pixmap_set_dri(struct sna *sna,
 	}
 
 	assert(priv->flush == false);
+	assert(priv->cow == NULL);
 	assert(priv->cpu_damage == NULL);
 	assert(priv->gpu_bo);
 	assert(priv->gpu_bo->proxy == NULL);
@@ -440,7 +443,7 @@ static void _sna_dri_destroy_buffer(struct sna *sna, DRI2Buffer2Ptr buffer)
 		     pixmap->drawable.serialNumber,
 		     pixmap == sna->front));
 
-		list_del(&priv->list);
+		list_del(&priv->flush_list);
 
 		priv->gpu_bo->flush = false;
 		priv->pinned &= ~PIN_DRI;
@@ -506,6 +509,9 @@ static void set_bo(PixmapPtr pixmap, struct kgem_bo *bo)
 	assert((priv->pinned & PIN_PRIME) == 0);
 	assert(priv->flush);
 
+	if (priv->cow)
+		sna_pixmap_undo_cow(sna, priv, 0);
+
 	/* Post damage on the new front buffer so that listeners, such
 	 * as DisplayLink know take a copy and shove it over the USB,
 	 * also for software cursors and the like.
@@ -520,12 +526,13 @@ static void set_bo(PixmapPtr pixmap, struct kgem_bo *bo)
 		       pixmap->drawable.width,
 		       pixmap->drawable.height);
 	sna_damage_destroy(&priv->cpu_damage);
-	list_del(&priv->list);
+	list_del(&priv->flush_list);
 	priv->cpu = false;
 
 	assert(bo->refcnt);
 	if (priv->gpu_bo != bo) {
-		kgem_bo_destroy(&sna->kgem, priv->gpu_bo);
+		if (priv->gpu_bo)
+			kgem_bo_destroy(&sna->kgem, priv->gpu_bo);
 		priv->gpu_bo = ref(bo);
 		if (priv->mapped) {
 			assert(!priv->shm && priv->stride);
@@ -872,10 +879,10 @@ sna_dri_copy(struct sna *sna, DrawablePtr draw, RegionPtr region,
 }
 
 static bool
-can_blit(struct sna * sna,
+can_blit(struct sna *sna,
 	 DrawablePtr draw,
-	 DRI2BufferPtr front,
-	 DRI2BufferPtr back)
+	 DRI2BufferPtr dst,
+	 DRI2BufferPtr src)
 {
 	RegionPtr clip;
 	int w, h;
@@ -884,28 +891,34 @@ can_blit(struct sna * sna,
 	if (draw->type == DRAWABLE_PIXMAP)
 		return true;
 
-	if (get_private(front)->pixmap != get_drawable_pixmap(draw)) {
-		DBG(("%s: reject as front pixmap=%ld, but expecting pixmap=%ld\n",
+#if 0
+	if (get_private(dst)->pixmap != get_drawable_pixmap(draw)) {
+		DBG(("%s: reject as dst pixmap=%ld, but expecting pixmap=%ld\n",
 		     __FUNCTION__,
-		     get_private(front)->pixmap ? get_private(front)->pixmap->drawable.serialNumber : 0,
+		     get_private(dst)->pixmap ? get_private(dst)->pixmap->drawable.serialNumber : 0,
 		     get_drawable_pixmap(draw)->drawable.serialNumber));
 		return false;
 	}
 
+	assert(sna_pixmap(get_private(dst)->pixmap)->flush);
+#endif
+	assert(get_private(dst)->bo->flush);
+	assert(get_private(src)->bo->flush);
+
 	clip = &((WindowPtr)draw)->clipList;
 	w = clip->extents.x2 - draw->x;
 	h = clip->extents.y2 - draw->y;
 	if ((w|h) < 0)
 		return false;
 
-	s = get_private(front)->size;
+	s = get_private(dst)->size;
 	if ((s>>16) < h || (s&0xffff) < w) {
 		DBG(("%s: reject front size (%dx%d) < (%dx%d)\n", __func__,
 		       s&0xffff, s>>16, w, h));
 		return false;
 	}
 
-	s = get_private(back)->size;
+	s = get_private(src)->size;
 	if ((s>>16) < h || (s&0xffff) < w) {
 		DBG(("%s:reject back size (%dx%d) < (%dx%d)\n", __func__,
 		     s&0xffff, s>>16, w, h));
@@ -945,8 +958,6 @@ sna_dri_copy_region(DrawablePtr draw,
 	if (!can_blit(sna, draw, dst_buffer, src_buffer))
 		return;
 
-	assert(sna_pixmap(pixmap)->flush);
-
 	if (dst_buffer->attachment == DRI2BufferFrontLeft) {
 		dst = sna_pixmap_get_bo(pixmap);
 		copy = (void *)sna_dri_copy_to_front;
@@ -2212,6 +2223,16 @@ sna_dri_async_swap(ClientPtr client, DrawablePtr draw,
 }
 #endif
 
+static uint64_t gettime_us(void)
+{
+	struct timespec tv;
+
+	if (clock_gettime(CLOCK_MONOTONIC, &tv))
+		return 0;
+
+	return (uint64_t)tv.tv_sec * 1000000 + tv.tv_nsec / 1000;
+}
+
 /*
  * Get current frame count and frame count timestamp, based on drawable's
  * crtc.
@@ -2223,13 +2244,16 @@ sna_dri_get_msc(DrawablePtr draw, CARD64 *ust, CARD64 *msc)
 	drmVBlank vbl;
 	int pipe;
 
-	/* Drawable not displayed, make up a value */
-	*ust = *msc = 0;
 
 	pipe = sna_dri_get_pipe(draw);
 	DBG(("%s(pipe=%d)\n", __FUNCTION__, pipe));
-	if (pipe == -1)
+	if (pipe == -1) {
+fail:
+		/* Drawable not displayed, make up a *monotonic* value */
+		*ust = gettime_us();
+		*msc = 0;
 		return TRUE;
+	}
 
 	VG_CLEAR(vbl);
 	vbl.request.type = DRM_VBLANK_RELATIVE | pipe_select(pipe);
@@ -2242,6 +2266,7 @@ sna_dri_get_msc(DrawablePtr draw, CARD64 *ust, CARD64 *msc)
 	} else {
 		DBG(("%s: query failed on pipe %d, ret=%d\n",
 		     __FUNCTION__, pipe, errno));
+		goto fail;
 	}
 
 	return TRUE;
@@ -2362,13 +2387,24 @@ out_complete:
 }
 #endif
 
+static bool has_i830_dri(void)
+{
+	return access(DRI_DRIVER_PATH "/i830_dri.so", R_OK) == 0;
+}
+
 static const char *dri_driver_name(struct sna *sna)
 {
 	const char *s = xf86GetOptValString(sna->Options, OPTION_DRI);
 	Bool dummy;
 
-	if (s == NULL || xf86getBoolValue(&dummy, s))
-		return sna->kgem.gen < 040 ? "i915" : "i965";
+	if (s == NULL || xf86getBoolValue(&dummy, s)) {
+		if (sna->kgem.gen < 030)
+			return has_i830_dri() ? "i830" : "i915";
+		else if (sna->kgem.gen < 040)
+			return "i915";
+		else
+			return "i965";
+	}
 
 	return s;
 }
